/*  reader-writer-log.c (Praktikum, Arbeitsblatt 9)

    Vorlesung Betriebssysteme
    Hans-Georg Esser, Hochschule Muenchen
    hans-georg.esser@hm.edu

    Musterloesung, mit Log-Datei, v1.0 (2011/07/06)

    Uebersetzen mit: 
    gcc -o reader-writer reader-writer.c -lpthread

*/

#include <pthread.h>     // POSIX-Threads
#include <semaphore.h>   // POSIX-Semaphore
#include <stdio.h>       // printf, stdout

#define N 20             // Groesse des Puffers
#define NO_READERS 3
#define NO_WRITERS 3
#define NOCHAR '_'       // "_" bedeutet: leere Pufferposition
#define LOGFILE "output.dat"   // Log-File

FILE *f;                 // globale Variable fuer Logdatei

int new_thread;

int consumerids[NO_READERS];


// Semaphor-Definitionen, siehe: man sem_overview
static sem_t mutex;      // kontrolliert Zugriff auf Puffer
static sem_t empty;      // zaehlt freie   Plaetze im Puffer
static sem_t full;       // zaehlt belegte Plaetze im Puffer

unsigned char buffer[N+1]; // Puffer
int next_read;           // naechste Leseposition
int next_write;          // naechste Schreibposition

// Nachricht, die ueber den Puffer uebertragen wird
// (mit "_" terminiert)
unsigned char inmessage[]="Das ist die Eingangsnachricht, die ueber den Puffer uebertragen werden soll._";
int inmessage_pos = 0;
int inmessage_len = sizeof(inmessage);

// Zeichen aus Puffer auslesen
unsigned char remove_item() {
  unsigned char c;
  c = buffer[next_read++];
  next_read %= N;        // Position inkrementieren, modulo N
  return c;
}

// Zeichen "produzieren", also fuer Eintragen in den Puffer
// vorbereiten
unsigned char produce_item() { 
  if (inmessage_pos < inmessage_len) {
    return inmessage[inmessage_pos++];
  } else {
    return NOCHAR;
  }
}

// Zeichen an aktueller Schreibposition in den Puffer schreiben
// und Schreibposition erhoehen (Puffer ist Ring!)
void enter_item(unsigned char item) {
  buffer[next_write++] = item;
  next_write %= N;         // modulo N !
}

// Zeichen "verbrauchen": einfach ausgeben. Die Nachricht
// erscheint also nach dem Umweg ueber den Puffer auf dem
// Bildschirm
void consume_entry(unsigned char item) { 
  if (item != NOCHAR) {
    write(1,&item,1); fflush(stdout);
  }
};

// Puffer initialisieren: alle Zeichen auf "_" setzen
init_buffer() {
  int i;
  for (i=0; i<N; i++) { 
    buffer[i] = NOCHAR; 
  };
  next_read=0;
  next_write=0;
}


// Hauptprogramm des Producer-Prozesses -- wird als neuer
// Thread gestartet
void producer() {
  unsigned char item;
  while (item != NOCHAR) {
    // DEBUG: printf ("+"); fflush(stdout);
    usleep(5);
    // item = produce_item();// Erzeuge etwas fr den Puffer
    sem_wait (&empty);    // Leere Plaetze dekrementieren bzw. blockieren
    sem_wait (&mutex);    // Eintritt in den kritischen Bereich
    item = produce_item();// Erzeuge etwas fr den Puffer
    enter_item (item);    // In den Puffer einstellen
    sem_post (&mutex);    // Kritischen Bereich verlassen
    sem_post (&full);     // Belegte Plaetze erhoehen, evtl. Consumer wecken
  }
}

// Hauptprogramm des Consumer-Prozesses -- wird als neuer
// Thread gestartet
void consumer(int * id) {
  unsigned char item;
  int tid = (int) (*id);  // Thread-ID rausfinden
  while (item != NOCHAR) {
    usleep(5);
    // DEBUG: printf ("-"); fflush(stdout);
    sem_wait (&full);       // Belegte Plaetze dekrementieren bzw. blockieren
    sem_wait (&mutex);      // Eintritt in den kritischen Bereich
    item = remove_item();   // Aus dem Puffer entnehmen
    consume_entry (item);   // Verbrauchen
    sem_post (&mutex);      // Kritischen Bereich verlassen
    sem_post (&empty);      // Freie Plaetze erhoehen, evtl. Producer wecken
    // consume_entry (item);   // Verbrauchen
    list(tid);              // beim Verbrauchen list() aufrufen
  }
}

// Status von Puffer und Semaphoren ausgeben
list(tid) {
  int i,v_mutex, v_empty, v_full;
  unsigned char ch;
  sem_getvalue (&mutex, &v_mutex);
  sem_getvalue (&empty, &v_empty);
  sem_getvalue (&full,  &v_full);
  // printf ("Buffer: [%s]\n", buffer);
  /* for (i=0; i<N; i++) {
    ch = buffer[i];
    if ( ((int)ch<32) || ((int)ch>127) )  ch='#';
    write(1, &ch, 1);
  };
  */

  // Ausgabe: mutex, empty, full, next_read, next_write, buffer
  fprintf (f, "%2d | %2d | %2d | %2d | %2d | %2d | [%s] \n",
              (unsigned int) tid, 
              v_mutex, v_empty, v_full, next_read, next_write, buffer);
}

// Hauptprogramm
main() {
  int j;  // Counter fuer Logfile-Loop
  f = fopen(LOGFILE, "w");
  pthread_t consumer_thread[NO_READERS];  // Thread-Variablen fuer Consumer ...
  pthread_t producer_thread[NO_WRITERS];  // ... und Producer
  sem_init(&mutex, 0, 1);     // Init: 1, Wertebereich: 0-1 (Mutex)
  sem_init(&empty, 0, N);     // Init: N, Wertebereich: 0-N
  sem_init(&full,  0, 0);     // Init: 0, Wertebereich: 0-N

  init_buffer();
  fprintf (f, "Startwerte:\n");  
  list(0);
  fprintf (f, "\n\nAusfuehrung beginnt...\n\n");

  // Ausgabe: mutex, empty, full, next_read, next_write, buffer
  fprintf (f, "Log-Eintraege: \n");
  fprintf (f, "#1 = mutex \n");
  fprintf (f, "#2 = empty \n");
  fprintf (f, "#3 = full \n");
  fprintf (f, "#4 = next_read \n");
  fprintf (f, "#5 = next_write \n");
  
  fprintf (f, "ID | #1 | #2 | #3 | #4 | #5 | Puffer \n");
  fprintf (f, "---+----+----+----+----+----+--------------- \n");
  

  // Threads erzeugen
  int i;
  for ( i=0; i < NO_WRITERS; i++ ) {
    pthread_create( &producer_thread[i], NULL, (void*)&producer, NULL);
  }
  usleep(100);
  for ( i=0; i < NO_READERS; i++ ) {
    consumerids[i] = i+1;   // Thread-ID 1..NO_READERS
    pthread_create( &consumer_thread[i], NULL, (void*)&consumer, &consumerids[i]);
  }

  // Logfile Loop
  /*
  for (j=0; j<100; j++) {
    list();
    usleep(10);
  }
  */
  
  // Aufraeumen:
  for ( i=0; i < NO_WRITERS; i++ ) {
    pthread_join( producer_thread[i], NULL );
  }
  for ( i=0; i < NO_READERS; i++ ) {
    pthread_join( consumer_thread[i], NULL );
  }
  fclose (f);  // Logfile schliessen
  printf ("\n");
}
